# -*- coding: utf-8 -*-
from typing import List
from types import MethodType
from ._nodered import Node
from ._property import Property
from ._server import Server


def register(name:str, category:str = "nodered_py", properties:List[Property] = []) -> MethodType:
    """
    Decorator to register Node function

    Parameters
    ----------
    name: str, required
        name of Node to register
    category: str, default nodered_py
        category of Node
    properties: List[noderedpy._property.Property]
        propertis of Node
    """
    def decorator(node_func:MethodType):
        node = Node(name if name.startswith("nodered-py") else f"nodered-py-{name}", category, properties, node_func)
        Server.registered_nodes.append(node)

        return node_func
    
    return decorator



# -*- coding: utf-8 -*-
import json
from typing import Any, Union, List



class Property:
    def __init__(self, name:str, default:Any = None, required:bool = False, display_icon:str = None):
        self.name, self.default, self.required, self.display_icon =\
            name, default, required, display_icon
        
    @property
    def display_name(self) -> str:
        return " ".join([
            item.capitalize()
            for item in self.name.split("_")
        ])

class CodeProperty(Property):
    def __init__(self, name:str, default:str = "", language:str = None, height:int = 250, required:bool = False, display_icon:str = None):
        if not isinstance(default, str):
            raise TypeError("CodeProperty can accept types: [ 'dict', 'json string' ]")
        
        super().__init__(name, default, required, display_icon if display_icon else "fa fa-code")
        self.language, self.height =\
            language, height

class InputProperty(Property):
    def __init__(self, name:str, default:Union[int, float, str] = None, required:bool = False, display_icon:str = None):
        if not isinstance(default, (int, float, str)):
            raise TypeError("InputProperty can accept types: [ 'int', 'float', 'str' ]")
        
        if default is None:
            if isinstance(default, int):
                default = 0
            elif isinstance(default, float):
                default = 0.0
            else:
                default = ""

        super().__init__(name, default, required, display_icon)

        if display_icon is None:
            if isinstance(default, (int, float)):
                self.display_icon = "fa fa-sort-numeric-asc"
            else:
                self.display_icon = "fa fa-font"
        else:
            self.display_icon = display_icon

class ListProperty(Property):
    def __init__(self, name:str, default:list = [], height:int = 250, required:bool = False, display_icon:str = None):
        if not isinstance(default, list):
            raise TypeError("ListProperty can accept type: 'list'")

        super().__init__(name, default, required, display_icon if display_icon else "fa fa-list-ul")
        self.height = height

class DictProperty(CodeProperty):
    def __init__(self, name:str, default:Union[dict, str] = {}, height:int = 250, required:bool = False, display_icon:str = None):
        if not isinstance(default, ( dict, str )):
            raise TypeError("DictProperty can accept types: [ 'dict', 'json string' ]")
        
        if isinstance(default, str):
            if not default.strip().startswith("{") or not default.strip().endswith("}"):
                raise ValueError("DictProperty value must be 'dict' or 'json string'!")

            try:
                json.loads(default)
            except:
                raise ValueError("DictProperty value must be 'dict' or 'json string'!")

        if isinstance(default, dict):
            default = json.dumps(default, indent = 4)

        super().__init__(name, default, "json", height, required, display_icon if display_icon else "fa fa-code")

class SpinnerProperty(Property):
    def __init__(self, name:str, default:float = 0, step:float = None, min:float = None, max:float = None, required:bool = False, display_icon:str = None):
        if not isinstance(default, (int, float)):
            raise TypeError("SpinnerProperty can accept types: [ 'int', 'float' ]")

        super().__init__(name, default, required, display_icon if display_icon else "fa fa-random")
        self.step, self.min, self.max =\
            step, min, max

class ComboBoxProperty(Property):
    def __init__(self, name:str, items:List[Any], default:str = None, required:bool = False, display_icon:str = None):
        if not isinstance(items, list):
            raise TypeError("items of ComboBoxProperty must be type: 'list'")
        
        super().__init__(name, default, required, display_icon if display_icon else "fa fa-filter")
        self.items = items



# -*- coding: utf-8 -*-
import os, sys, json, shutil, asyncio
from glob import glob
from typing import List, Type
from types import MethodType
from aiohttp import web
from ._nodered import RED, Node
from ._property import Property


class Server(web.Application):
    """
    Server to communicate with Node-RED
    """
    registered_nodes:List[Node] = []

    def __init__(self, node_red:RED):
        """
        Initialize Server to communicate with Node-RED

        Parameters
        ----------
        node_red: RED, required
            Node-RED settings
        """
        super().__init__()

        self.__red = node_red
        # setup Node-RED
        node_red.start_for_ready()

    async def __call_node(self, req:web.Request, node_name:str):
        node = list(filter(lambda n: n.name == node_name, Server.registered_nodes))[0]

        data = await req.json()
        return web.Response(
            text = json.dumps(node.run(data["props"], data["msg"])),
            content_type = "application/json"
        )

    def __on_started(self):
        if self.__show_browser:
            import webbrowser
            webbrowser.open_new(f"http://127.0.0.1:{self.__red.port}/node-red")

        if self.__start_callback:
            self.__start_callback()


    def register(self, node_func:MethodType, name:str, category:str = "nodered_py", properties:List[Property] = []):
        """
        Function to register Node function

        Parameters
        ----------
        node_func: MethodType, required
            Node function to register
        name: str, required
            name of Node to register
        category: str, default nodered_py
            category of Node
        properties: List[noderedpy._property.Property]
            propertis of Node
        """
        node = Node(name if name.startswith("nodered-py") else f"nodered-py-{name}", category, properties, node_func)
        Server.registered_nodes.append(node)

    def start(self, port:int, show_browser:bool = True, callback:MethodType = None):
        """
        Start Server

        Parameters
        ----------
        port: int, required
            port for Server
        show_browser: bool, default True
            show browser when started or not
        callback: MethodType, default None
            callback when Server started
        """
        self.__port, self.__show_browser, self.__start_callback = port, show_browser, callback
        self.__red.start(callback = self.__on_started, server = Server)

        # remove existing nodes
        for node_dir in glob(os.path.join(self.__red.user_dir, "node_modules", "nodered-py-*")):
            shutil.rmtree(node_dir)

        # create nodes
        for node in Server.registered_nodes:
            node.create(self.__red.user_dir, port)

        # map nodes to route
        self.add_routes([
            web.get("", lambda _: web.HTTPFound(f"http://127.0.0.1:{self.__red.port}/node-red")),
            web.post("/nodes/{node_name}", lambda req: self.__call_node(req, req.match_info["node_name"]))
        ])

        # start server
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        server = loop.create_server(self.make_handler(), host = "127.0.0.1", port = port)
        loop.run_until_complete(server)
        loop.run_forever()

    def stop(self):
        """
        Stop Server
        """
        import psutil, signal

        self.__red.stop()

        killed = False
        for process in psutil.process_iter():
            try:
                for conns in process.connections(kind = "inet"):
                    if conns.laddr.port == self.__port:
                        process.send_signal(signal.SIGTERM)
                        killed = True
                        break
            except psutil.AccessDenied:
                pass

            if killed:
                break

class StandaloneServer:
    """
    Server to Communicate with Node-RED and webview
    """
    def __init__(self, node_red:RED):
        """
        Initialize Server to communicate with Node-RED

        Parameters
        ----------
        node_red: RED, required
            Node-RED settings
        """
        self.__server = Server(node_red)
        self.__red = node_red

    def register(self, node_func:MethodType, name:str, category:str = "nodered_py", properties:List[Property] = []):
        """
        Function to register Node function

        Parameters
        ----------
        node_func: MethodType, required
            Node function to register
        name: str, required
            name of Node to register
        category: str, default nodered_py
            category of Node
        properties: List[noderedpy._property.Property]
            propertis of Node
        """
        self.__server.register(node_func, name, category, properties)

    def start(self, title:str, width:int = 1000, height:int = 650, x:int = None, y:int = None, debug:bool = False, port:int = 1881, default_root:str = None):
        """
        Start Server and show webview

        Parameters
        ----------
        title: str, required
            title of webview window
        width: int, default 1000
            default width of webview window
        height: int, default 650
            default height of webview window
        x: int, default None
            default x location of webview window
        y: int, default None
            default y location of webview window
        debug: bool, default False
            debug(devtools) or not
        port: int, default 1881
            port for Server
        default_root: str, default None
            default url root to show
            if None, use admin_root of Node-RED settings
        """
        default_root = self.__red.admin_root if default_root is None else default_root
        if not default_root.startswith("/"):
            raise ValueError("default_root must startswith '/'!")

        import webview

        webview.initialize("cocoa" if sys.platform == "darwin" else "cef" if sys.platform == "win32" else "qt")
        win = webview.create_window(title, width = width, height = height, x = x, y = y)
        win.events.closing += self.__server.stop

        webview.start(lambda: self.__server.start(port, False, lambda: win.load_url(f"http://127.0.0.1:{self.__red.port}{default_root}")), debug = debug)



# -*- coding: utf-8 -*-
import os, sys, subprocess, threading, traceback, noderedpy
from types import MethodType
from typing import Type, List
from .templates import package_json, node_html, node_js
from ._property import Property
from . import __path__


class RED:
    """
    Node-RED manager class
    """
    def __init__(self, user_dir:str, admin_root:str, port:int = 1880, show_default_category:bool = True):
        """
        Set configs of Node-RED and setup

        Parameters
        ----------
        user_dir: str, required
            userDir of Node-RED settings
        admin_root: str, required
            httpAdminRoot of Node-RED settings
        port: int, default 1880
            port of Node-RED server
        show_default_category: bool, default True
            show default categories of Node-RED or not
        """
        self.user_dir, self.admin_root, self.port, self.show_default_category =\
            user_dir, admin_root, port, show_default_category

        # setup Node-RED starter
        subprocess.call(
            [ "npm", "install" ],
            stdout = subprocess.DEVNULL,
            cwd = os.path.join(__path__[0], "node-red-starter")
        )

    # default started callback
    def __on_started(self, callback:MethodType = None):
        while True:
            if os.path.exists(self.__started_file):
                break

        callback()

    def start(self, wait:bool = False, debug:bool = True, callback:MethodType = None, server:Type["noderedpy._server.Server"] = None):
        """
        Start Node-RED server

        Parameters
        ----------
        wait: bool, default False
            wait for Node-RED server process of not
        debug: bool, default True
            show outputs on console or not
        callback: MethodType, default None
            callback when Node-RED server started
        server: Type[noderedpy._server.Server], default None
            server for setup "user category"
        """

        # set started flag file
        self.__started_file = os.path.join(__path__[0], "node-red-starter", "started")
        # kill if process listen on port
        self.stop()

        # map callback
        if callback:
            threading.Thread(target = self.__on_started, args = (callback,), daemon = True).start()

        # set subprocess args
        args = [
            "node",
            "index.js",
            f"--user-dir={self.user_dir}",
            f"--admin-root={self.admin_root}",
            f"--port={self.port}",
            f"--show-default-category={'true' if self.show_default_category else 'false'}"
        ]
        if server:
            args.append(
                f"--user-category={','.join(list(set([ node.category for node in server.registered_nodes ])))}"
            )

        # run Node-RED server
        proc = subprocess.Popen(
            args, shell = False, stdout = sys.stdout if debug else subprocess.DEVNULL,
            cwd = os.path.join(__path__[0], "node-red-starter")
        )
        if wait:# wait if flag is True
            proc.wait()

    def start_for_ready(self):
        """
        Start Node-RED for setup default userDir
        """
        def on_started():
            os.remove(self.__started_file)
            self.__stop()

        self.start(True, False, on_started)

    def __stop(self):
        import psutil, signal

        killed = False
        for process in psutil.process_iter():
            try:
                for conns in process.connections(kind = "inet"):
                    if conns.laddr.port == self.port:
                        process.send_signal(signal.SIGTERM)
                        killed = True
                        break
            except psutil.AccessDenied:
                pass

            if killed:
                break

    def stop(self):
        """
        Stop Node-RED server
        """
        if os.path.exists(self.__started_file):
            os.remove(self.__started_file)

        self.__stop()

class Node:
    def __init__(self, name:str, category:str, properties:List[Property], node_func:MethodType):
        if " " in name.strip():
            raise NameError("Node name cannot contain spaces!")
        
        if "-" in category.strip() or "," in category.strip():
            raise NameError("Category cannot contain '-' or ','!")

        self.name, self.category, self.properties =\
            name, category, properties
        
        self.__node_func = node_func

    def create(self, node_red_user_dir:str, port:int):
        node_dir = os.path.join(node_red_user_dir, "node_modules", self.name)
        os.makedirs(os.path.join(node_dir, "lib"))

        with open(os.path.join(node_dir, "package.json"), "w", encoding = "utf-8") as pjw:
            pjw.write(package_json(self))

        with open(os.path.join(node_dir, "lib", f"{self.name}.html"), "w", encoding = "utf-8") as nhw:
            nhw.write(node_html(self))

        with open(os.path.join(node_dir, "lib", f"{self.name}.js"), "w", encoding = "utf-8") as njw:
            njw.write(node_js(self, port))

    def run(self, props:dict, msg:dict) -> dict:
        print(f"{self.name} started\n===================================")
        try:
            resp = self.__node_func(props, msg)
            print("============================= ended\n")

            resp.update({ "state": "success" })

            return resp
        except:
            print("============================= error\n")
            return { "state": "fail", "message": traceback.format_exc() }



# -*- coding: utf-8 -*-
from setuptools import setup
from noderedpy import __version__, __author__, __email__


with open("requirements.txt", "r", encoding = "utf-8") as reqr:
    requires = [
        item.strip()
        for item in reqr.read().split("\n")
    ]

with open("README.md", "r", encoding = "utf-8") as rmr:
    readme = rmr.read()


setup(
    # publish informations
    name = "nodered.py",
    author = __author__,
    author_email = __email__,
    url = "https://github.com/oyajiDev/NodeRED.py",
    version = __version__,
    python_requires = ">=3.9",
    install_requires = requires,
    setup_requires = requires,
    license = "MIT license",
    description = "make python function to Node-RED node",
    long_description = readme,
    long_description_content_type = "text/markdown",
    # package informations
    packages = [
        "noderedpy", "noderedpy/templates", "noderedpy/node-red-starter"
    ],
    package_data = {
        "": [
            "*.html", "*.js", "*.json"
        ]
    },
    include_package_data = True,
    zip_safe = True
)



# -*- coding: utf-8 -*-
from ._nodered import RED, Node
from ._property import (
    InputProperty, ListProperty, DictProperty,
    SpinnerProperty, ComboBoxProperty, CodeProperty
)
from ._server import Server, StandaloneServer


__version__ = "0.1.4"
__author__ = "oyajiDev"
__email__ = "this.dev.somehit@gmail.com"

__all__ = [
    "RED", "Node",
    "InputProperty", "ListProperty", "DictProperty",
    "SpinnerProperty", "ComboBoxProperty", "CodeProperty",
    "Server", "StandaloneServer"
]
